/**********************************************************************
 *
 * cfgread.cpp -- 
 * Copyright (C) 1999  The New Zealand Digital Library Project
 *
 * PUT COPYRIGHT NOTICE HERE
 *
 * $Id: cfgread.cpp 131 1999-02-02 09:54:26Z rjmcnab $
 *
 *********************************************************************/

/*
   $Log$
   Revision 1.4  1999/02/02 09:54:26  rjmcnab

   Changed the configuration file reader so that any space can be used to
   seperate items and so that quotes can be used to contain items. I also
   allowed a '\' to indicate that the line continues on the next line.

   Revision 1.3  1999/01/12 01:50:57  rjmcnab

   Standard header.

   Revision 1.2  1999/01/08 02:33:13  rjmcnab

   Added standard header to source files.

 */


#include "cfgread.h"


// returns 0 on success, -1 on failure
int read_cfg_line (ifstream &filein, text_tarray &values) {
  //  outconvertclass text_t2ascii;

  values.erase(values.begin(), values.end());

  if (!filein.good()) return -1;

  text_t curvalue;
  char c, quote;

  bool linecontinues = true;
  filein.get(c);
  while (!filein.eof() && linecontinues) {
    // eat up all return characters
    while (!filein.eof() && (c=='\n' || c=='\r')) { filein.get(c); }

    // ignore comments
    if (c == '#') {
      while (!filein.eof() && c!='\n' && c!='\r') { filein.get(c); }
    }

    // deal with all the records on this line
    while (!filein.eof() && c!='\n' && c!='\r') {
      // ignore white space
      while (!filein.eof() && (c==' ' || c=='\t')) { filein.get(c); }

      // get the next record
      if (c!='\n' && c!='\r') {
	curvalue.clear();
	quote = 0;

	// see if this is a quoted record
	if (c=='\'' || c=='\"') {
	  quote = c;
	  filein.get(c);
	}

	// get the value
	while (!filein.eof() && c!='\n' && c!='\r') {
	  if (quote != 0 && c == quote) { // end of quote
	    filein.get(c);
	    break;
	  } else if (quote == 0 && (c == ' ' || c == '\t')) {
	    break; // space seperator
	  }
	  curvalue.push_back(c);
	  filein.get(c);
	}

	// see if the record continues on the next line
	if ((c=='\n' || c=='\r') && curvalue=="\\") {
	  linecontinues = true;
	} else {
	  //	  cerr << text_t2ascii << "\"" << curvalue << "\" ";
	  
	  values.push_back(curvalue);
	  curvalue.clear();
	  linecontinues = false;
	}
      }
    }
  }

  //  cerr << "\n";

  return 0;
}
